####################
# version 2.0
####################


################################################################################
#                                                                              #
# NAME Bio.hit
# DESC Hit Object
#                                                                              #
################################################################################
"""
:NAME:
Bio.hit

:DESC:
Hit Object
Simple format: tuple
(position, strand, score, -log_P-value)
ex: (10, '+', 10.6, 3.0)

"""
# IMPORTS
#-------------------------------------------------------------------------------
from numpy import *

import Core.types
import Core.Types.array as Array
import psfm
import sequence
#-------------------------------------------------------------------------------
# IMPORTS


# OBJECTS
#-------------------------------------------------------------------------------

class Hits(Array.Array):
    """
    store hits for m matrices, s sequences
    hits for one sequence, one matrix is in tuple of array format or list of tuple
    h[m,s]
    """
    FULL = True

    def __init__(self, sequences, matrices):
        Array.Array.__init__(self, [[0 for i in range(len(sequences))] for j in range(len(matrices))])
        self.sequences = sequences
        self.matrices  = matrices


    def __resolve(self, key):
        i,j = key
        
        if type(i) != int and type(i) != slice:
            i = self.matrices.index(i)
        if type(j) != int and type(j) != slice:
            j = self.sequences.index(j)
        return i,j
        
    def __setitem__(self, key, value):
        """
        key   -- sequence, matrix or index,index (int,int)
        value -- hit for one sequence, one matrix
        """
        Array.Array.__setitem__(self, self.__resolve(key), value)

    def __getitem__(self, key):
        """
        key   -- sequence, matrix or int,int
        """
        return Array.Array.__getitem__(self, self.__resolve(key) )    

    #conversion
    def convert(self, format='array'):
        """convert hit format object for each matrix and sequence
        2 format: 
            -list of tuple 'list'
            -tuple of array 'array'  (array_of_position, array_of_score)
        """
        #for each
        #( array([i[0] for i in hits], 'i'), array([i[2] for i in hits], 'f') )
        #[ (hits[0][i], True, hits[1][i]) for i in range(len(hits[0])) ]
        raise NotImplementedError
    
    def count(self):
        S = 0
        for i in range(self.dim()[0]):
            for j in range(self.dim()[1]):
                S += len(Array.Array.__getitem__(self, (i,j) ))
        return S


   # def __str__(self, hideEmpty=True):
    #    """TXT representation of hits
     #   hideEmpty --  hide empty line (default=True)
      #  return string
      #  """
       # l =[]
        #header
        #l += ['%%{info} %d hits, %d sequences, %d matrices' % (self.count(), self.dim()[1], self.dim()[0])]
        #l += ['%{doc} for each sequence each matrix : list of hits (position, score)']
    
        #for s in range(len(self.sequences)):
         #   l.append('>%s' % `self.sequences[s]`)
          #  for m in range(len(self.matrices)):
           #     if hideEmpty and len(self[m,s]) == 0:
            #        continue
             #   shift = self.sequences[s].location[0]
              #  if len(self[m,s]) > 0 and len(self[m,s][0]) > 1 and Hits.FULL:
               #     l += ['\t%s ' % (self.matrices[m].id) + ' '.join( [ '%-5d %.2f' % (shift+h[0], h[2]) for h in self[m,s]] )]
                #else:
                 #   l += ['\t%s ' % (self.matrices[m].id) + ' '.join( [ '%-5d ' % (shift+h[0]) for h in self[m,s]] )]
        #return '\n'.join(l)



    def __str__(self, hideEmpty=True):
        """TXT representation of hits

        hideEmpty --  hide empty line (default=True)
        return string
        """
        l =[]
        #header
        l += ['%%{info} %d hits, %d sequences, %d matrices' % (self.count(), self.dim()[1], self.dim()[0])]
        l += ['%{doc} for each sequence each matrix : list of hits (position, score)']
    
        for m in range(len(self.matrices)):
            l.append('>%s' % `self.matrices[m]`)
	    l.append(str(len(self.matrices[m])))
            for s in range(len(self.sequences)):
                if hideEmpty and len(self[m,s]) == 0:
                    continue
                shift = self.sequences[s].location[0]
                if len(self[m,s]) > 0 and len(self[m,s][0]) > 1 and Hits.FULL:
                    l += ['\t%s ' % (self.sequences[s].id) + ' '.join( [ '%-5d %s %.2f' % (shift+h[0],h[1], h[2]) for h in self[m,s]] )]
                else:
                    l += ['\t%s ' % (self.sequences[s].id) + ' '.join( [ '%-5d ' % (shift+h[0]) for h in self[m,s]] )]
        return '\n'.join(l)




    __repr__ = __str__


    def overlapping(self, m1, m2, location):
        """
        return  (overlap1, count1), (overlap2, count2)
        """
        h1 = self.selection2formatedlist(self.select(m1, location))
        h2 = self.selection2formatedlist(self.select(m2, location))
        return overlapping(h1, len(m1), h2, len(m2), location)

    def select(self, matrix, location):
        """
        return hits(list of list of tuple) selected (in [i,j]) for given matrix
        """
        i,j = location
        return hits_in_sequences(self.sequences, self[matrix,:], i, j)


    def selection2formatedlist(self, l):
        """
	    return a list of hits in a specific order
        """
        r = []
        for i in range(len(l)):
            r.append( [ (self.sequences[i].id, h[0], h[0] + self.sequences[i].location[0], h[1], h[2]) for h in l[i]] )
        return r

    def hits_in(self, matrix, location):
	"""
	return the number of hits for a matrix in a location
	"""
        i,j = location
        return [ len(h) for h in self.select(matrix, location)]


    def freq(self, matrix, location):
	"""
	return the frequency of occurency of a matrix in a location
	"""
        N = self.count_hits_in(matrix, location)
        #L = len(self.sequences) * (min(location[1], self.location[1] - len(matrix) + 1) - location[0] + 1)
        L = 0
        for s in self.sequences:
            L += min(location[1], s.location[1] - len(matrix) + 1) - location[0] + 1
        #print L
        return N / float(L)

    def count_hits_in(self, matrix, location):
	"""
	return the number of hits for a matrix in a location
	"""
        return Core.types.list.sum( [ len(h) for h in self.select(matrix, location)] )
        
    
    def sequences_with_hits_in(self, matrix, location):
	"""
	return a vector with the number of hits for each sequence
	[0,2,1,5,0]
	"""
        l = []
        x = -1
        i,j = location
        for hits in self[matrix,:]:
            x += 1
            for h in hits: #for each hit
                if h[0] < i - self.sequences[x].location[0]:
                    continue
                if h[0] > j - self.sequences[x].location[0]:
                    break
                l.append(self.sequences[x])
                break
        return l


    def sequences_vector(self, matrix, location):
        """
        sequences with hit in
        [0,1,0..]
        """
        l = [0] * len(self.sequences)
        x = -1
        i,j = location
        
        for s in range(len(self.sequences)):

            hits = self[matrix,s]
            for h in hits: #for each hit
                if h[0] < i - self.sequences[s].location[0]:
                    continue
                if h[0] > j - self.sequences[s].location[0]:
                    break
                l[s] = 1
                break
        return l


    #def percent_sequences_with_hits_in(self, matrix, location):
    #    return len(self.sequences_with_hits_in(matrix, location)) / float(len(self.sequences))



#-------------------------------------------------------------------------------
# OBJECTS

# CODE
#-------------------------------------------------------------------------------

def select_region(h, region):
    """
    h1 -- list of list of tuple (sequences, hit, (seqname, ....)
    """
    hr = []
    for s in range(len(h)):
        hr.append([])
        for i in range(len(h[s])):
            p = h[s][i][2]
            if p < region[0]:
                continue
            if p > region[1]:
                break

            hr[s].append(h[s][i])
    return hr




def overlapping(h1, ml1, h2, ml2, region):
    """
    h1 -- list of list of tuple (sequences, hit, (seqname, ....)
    (returned by selection2formatedlist
    ml1 -- len of matrix 1
    h2 -- list of list of tuple (sequences, hit, 
    ml2 -- len of matrix 2

    region -- tuple (a,b) 

    return (overlap1, count1), (overlap2, count2)
    """
    overlapCount1 = {}
    count1 = 0
    overlapCount2 = {}
    count2 = 0


    #select hit in region
    #print region
    h1 = select_region(h1, region)
    h2 = select_region(h2, region)


    for s in range(len(h1)):
        i1 = 0
        i2 = 0
        count1 += len(h1[s])
        count2 += len(h2[s])

        while i1 < len(h1[s]) and i2 < len(h2[s]): 

            #print i1, i2
            p1 = h1[s][i1][2]
            p2 = h2[s][i2][2]

            if (p1 >= p2 and p1 < p2 + ml2) or (p2 >= p1 and p2 < p1 + ml1):
                #print p1, p2
                overlapCount1.setdefault((s, p1), 1)
                overlapCount2.setdefault((s, p2), 1)            

                if p1 + ml1 < p2 + ml2:
                    i1 += 1
                else:
                    i2 += 1

            else:
                if p1 <= p2:
                    i1 += 1
                else:
                    i2 += 1

    return (len(overlapCount1), count1), (len(overlapCount2), count2)


def location2indicies(location, genelocation):
    """Convert location to list indicies

    location     -- feature location in REF coordinate
    genelocation -- gene location in REF coordinate
    """
    return (location[0] - genelocation[0], location[1] - genelocation[0]+1)

def location2slice(location, genelocation):
    """Convert location to list slice

    location     -- feature location in REF coordinate
    genelocation -- gene location in REF coordinate
    """
    return slice(location[0] - genelocation[0], location[1] - genelocation[0]+1)



def hits_in_sequences(sequences, lhits, i, j):
    """find hits in [i,j] for one sequence
    sequences -- list of Sequence
    lhits     -- list of hit (tuple)
    i, j      -- relative positions
    
    return list of hits in [i,j] (for each sequence)
    """
    assert(len(sequences) == len(lhits))
    l = []
    for x in range(len(sequences)): #for each sequence
        l.append(hits_in_sequence(sequences[x], lhits[x], i, j))
    return l

def hits_in_sequence(sequence, hits, i, j):
    """find hits in [i,j] for one sequence
    sequence -- Sequence.sequence.sequence
    hits     -- list of tuple
    i, j     -- relative positions
    
    return list of hits in [i,j]
    """
    l = []
    for h in hits: #for each hit
        if h[0] < i - sequence.location[0]:
            continue
        if h[0] > j - sequence.location[0]:
            break
        l.append(h)
    return l



def find_location(locations):
    """Find min and max of list of location
    """
    return (min([i[0] for i in locations]), max([i[1] for i in locations]))


def expand(hits, location, newLocation=None, binarize=True):
    """Expand hits in an array

    hits        -- (1 sequence, 1 matrix) list of tuples
    location    -- hits location
    newLocation -- fit hits in this new location
    Return array
    """
    positions = [h[0] for h in hits]
    if binarize:
        scores    = [1 for h in hits]
        type = 'i'
    else:
        scores    = [h[2] for h in hits]
        type = 'f'
    newLocation = newLocation or location
    tab = zeros(newLocation[1] - newLocation[0] + 1, type)
    put(tab[location[0] - newLocation[0]:location[0] - newLocation[0] + len(tab)], positions, scores)
    return tab

def expand_many2(hits, location, newLocation):
    """
    hits -- hits for one matrix n sequences
    """
    type = 'i'
    tab = zeros(newLocation[1] - newLocation[0] + 1, type)
    for h in hits:
        positions = [i[0] for i in h]
        scores    = [1 for i in h]
	t = zeros(newLocation[1] - newLocation[0] + 1, type)
        put(t[location[0] - newLocation[0]:location[0] - newLocation[0] + len(tab)], positions, scores)
	tab += t
    return tab

def expand_many(hits, location):
    """
    hits -- hits for one matrix n sequences
    """
    tab = zeros(location[1] - location[0] + 1, 'i')
    for h in hits:
	for i in h:
	    tab[i[0]] += 1
    return tab


def resize(a, location, tabLocation):
    """
    location must be included in tabLocation
    """
    tab = zeros(tabLocation[1] - tabLocation[0] + 1, 'i') -1
    tab[location[0] - tabLocation[0]:location[0] - tabLocation[0] + len(a)] = a
    return tab



#-------------------------------------------------------------------------------
# CODE




