####################
# version 2.0
####################


################################################################################
#                                                                              #
# NAME Bio.Types.PSFM  
# DESC PSFM Object
#                                                                              #
################################################################################
"""
:NAME:
Bio.Types.psfm

:DESC:
PSFM Objects

:LONG_DESC:
Position Frequency Matrices tools
  -Load from TRANSFAC file
  -Save to Patser format
  -...

TRANSFAC Matrices Ids:
example: V$NFKB_C
  -V$:vertebrates
  -I$: insects
  -P$: plants
  -F$: fungi
  -N$: nematodes
  -B$: bacteria


family:
'NF-kappaB'
'c-Rel'

[V$NFKB_Q6_01, V$NFKB_C, V$NFKB_Q6, V$NFKAPPAB_01, V$CREL_01]

classes:
    PSFMS
    PSFM
    
methods:
    transfac2dump
    jaspar2dump
    perso2dump
    read_matrix_file
    decode_matrix_title
"""

# IMPORTS
#-------------------------------------------------------------------------------
from copy import copy, deepcopy

import os
import sys
import re
import csv
from math import log

import numpy
import Core.IO.section
import Core.types as types
import Core.UI.cli as cli
import Core.Persistence.op as op

import transfac
import jaspar
import fasta
import input_checker as ic
#-------------------------------------------------------------------------------
# IMPORTS


# OBJECTS
#-------------------------------------------------------------------------------
def transfac2dump(datfilename, dumpfilename):
    """
    Create a binary file with a serialized object containing a list of pwm matrices
    in transfac format (from .dat file)
    
    datfilename  -- path to the .dat file containing the matrices
    
    dumpfilename -- the file to create
    """
    p = op.P(dumpfilename, mode='c')
    #p={}
    it = Core.IO.section.SectionIterator(open(datfilename))
    l = it.next() # first section contains version info
    
    p['_version'] = l[2:]
    p['_info'] = 'TRANSFAC MATRICES'
    accidx = {}
    size = 0
    info = cli.Info(length=None,showTime=False, showRemaining=False,verbosity=1)
    
    for x in it:

        matrix = PSFM.new_from_TRANSFAC_string(x)
	
        p[matrix.id] = matrix        
        accidx[matrix.acc] = matrix
        size += 1
        info('Building dump for %s ' % (matrix.id))
        
    p['_size'] = size
    p['!acc'] = accidx
    p.keys()
    #op.osave(p,dumpfilename)

def jaspar2dump(dir, dumpfilename):
    """
    Create a binary file with a serialized object containing a list of pwm matrices
    in jaspar format (list of pfm files)
    
    dir          -- path to the directory where the .pfm files and the annotation file are
    
    dumpfilename -- the file to create
    """
    p = op.P(dumpfilename, mode='c')
   # p={}
    matrices = jaspar.read_matrices(os.path.join(dir,'matrix_list.txt'), dir)
    nameidx = {}

    info = cli.Info(length=len(matrices),showTime=False, showRemaining=False,verbosity=1)

    for id in matrices:
                    
        matrix = PSFM.new_from_JASPAR(id, matrices[id])

        p[id] = matrix
        nameidx[matrix.id] = matrix
        info('Building dump for %s ' % (id))
        
    p['!name'] = nameidx
    p['_info'] = 'JASPAR MATRICES'
    p['_size'] = len(matrices)
    p.keys()
   # op.osave(p,dumpfilename)


def perso2dump(inputFile,dumpfilename):
    """    
    Create a binary file with a serialized object containing a list of pwm matrices
    given in a fixed format file
	
    inputFile    -- text file with the list of matrices. The format is strict:	
                   >ID,Name,Tax_group,Acc
                   matrice
                   
    dumpfilename -- binary file to create with the matrices
	
    """
    p = op.P(dumpfilename, mode='c')
	#p={}
    annotations={}
    matrices = {}
    nameidx = {}

    for title, mat in read_matrix_file(inputFile):
        id,name,tax,acc=decode_matrix_title(title)
		
        features=annotations.setdefault(id,{})
        features["name"]=name
        features["tax_group"]=tax
        features["acc"]=acc
        matrices[id]=dict(matrix=mat, features=annotations[id])
	
	info = cli.Info(length=len(matrices),showTime=False, showRemaining=False,verbosity=1)
    for id in matrices:
        # use Jaspar method because the format of the matrices is the same
        matrix=PSFM.new_from_JASPAR(id,matrices[id]) 
        p[id]=matrix
        nameidx[matrix.id]=matrix
        info('Building dump for %s ' % (id))
	    
    p['!name'] = nameidx
    p['_info'] = 'PERSO MATRICES'
    p['_size'] = len(matrices)
    p.keys()
	#op.osave(p,dumpfilename)
  
		

######################################################################
class PSFMS(list):
    def __init__(self, l, name=''):
        super(PSFM, self).__init__(l)
        self.name = ''

    def __repr__(self):
        return '%s (%d)' % (self.name, len(self))


class PSFM:
    """
    Position Specific Frequency matrix

    """
    def __init__(self, id, tab, orientation=1):
        """
        Position frequency TRANSFAC like matrix

        -tab: tab [[ACGT], [ACGT], ...] default orientation
          i: position
          j: base [A,C,G,T]
          value (frequency 0.0<=f<=1.0)
          
        -id: matrix identifier: string
        
        -orientation: 1 or 0: set it to 0 if orientation is [[1,2,3,4,5,6]#A, [1,2,3,4,5,6]#C, ...]

        """
        try:
            tab = [map(float, l) for l in tab]
            
            self.id      = id
            self.matrix  = numpy.array(tab, 'f') # i(line): position j: base
            
            if orientation == 0:
                self.matrix = numpy.transpose(self.matrix)

           
            self.ba      = 1.0 # Number of sequences used in the alignment 
            self.acc     = ''
            self.family  = ''
            self.bases   = ['A', 'C', 'G', 'T']
            self.desc    = '' #description
            self.version = ''
            

            self.renorm()
            self.gc   = self.calc_gc()
            self.info = self.info_content()
            
        except:
            cli.error('An input is malformed: '+id)
            sys.exit(1)


    def __len__(self):
        return self.matrix.shape[0]

    def __repr__(self):
        return '%s' % (self.id)

    def __str__(self):
        return self.to_str_format()
        #return '%-16s, %-12s' % (self.id, self.family)
        
    def __cmp__(self, other):
        if self.id == other.id and self.version == other.version:
            return 0
        elif self.id > other.id:
            return 1
        else:
            return -1

    def calc_gc(self):
        """Return GC% for the matrix
        """
        sum = 0.0

        for i in range(self.matrix.shape[0]):
            sum += float(numpy.add.reduce(self.matrix[i,1:3])) / numpy.add.reduce(self.matrix[i,:])
         
        return sum / self.matrix.shape[0]

    def reversed(self):
        """
        return a new reversed matrix
        """
        m = deepcopy(self)
        for j in range(4):
            m.matrix[:,4-j-1] = types.list.reversed(self.matrix[:,j].tolist())
        return m
	
    #
    # INPUT new matrix from string in TRANSFAC format
    #
    def new_from_TRANSFAC_string(string):
        """
	    return a PSFM object from a transfac matrix
        """
        r = transfac.read_matrix(string)
        m = PSFM(r['ID'], r['matrix'])
        m.ba = r['BA']
        m.bf = r['BF']
        m.acc = r['AC']
        m.family = r['NA']
        m.bases  = r['bases']
        m.species = r['species']
        
        return m

    new_from_TRANSFAC_string = staticmethod(new_from_TRANSFAC_string)

	###############################################################
 	   # modified by laurie
	############################"
    def new_from_JASPAR(id, jm):
        """
        return a PSFM object from a jaspar matrix
        """
        m = PSFM(id, jm['matrix'], orientation=0)
        m.acc = jm['features']['acc']
        m.family = jm['features']['name']
        m.group = jm['features']['tax_group']
        m.id  = id
        
        return m
    
    new_from_JASPAR = staticmethod(new_from_JASPAR)


    #
    # INPUT
    #
    def from_string(cls, string, id):
        """
        One line per base, one column per position

        """
        tab = [l.split() for l in string.splitlines()]
        pfm = PSFM(id, tab, 0)
        return pfm
    from_string = classmethod(from_string)

    def from_csv(cls, filename):
        """
        One line per base, one column per position

        """
        p = csv.parser()
        tab = [p.parse(l) for l in open(filename)]
        pfm = PSFM(os.path.basename(filename), tab, 0)
        return pfm
    
    from_csv = classmethod(from_csv)


    def test_ba(self):
        """
        Transform count matrix to frequency matrix
        """
        I, J = self.matrix.shape
        sum1 = 0
        for i in range(I):
            sum = round(numpy.add.reduce(self.matrix[i,:]), 1)

            if i > 0 and sum != sum1 and self.id[0:2] == 'V$':
                #print sum, sum1
                print self.acc
                break
            sum1 = sum

    def renorm(self):
        """
        Transform count matrix to frequency matrix
        """
        I, J = self.matrix.shape

        for i in range(I):
            sum = float(numpy.add.reduce(self.matrix[i,:]))
            
            # set  ba
            if self.ba == 1.0:
                self.ba = sum
                
            for j in range(J):
                self.matrix[i,j] = self.matrix[i,j] / sum
                if str(self.matrix[i,j]) == str(float('NaN')) :
                    cli.warning('a matrix is malformed (probably frequency null for all bases)')

    def renormed(self, m):
        """
        Transform count matrix to frequency matrix
        
        m      -- a count matrix to transform
        
        return -- a frequency matrix
        """
        m = m.copy()
        I, J = m.shape

        for i in range(I):
            sum = float(numpy.add.reduce(m[i,:]))
            for j in range(J):
                m[i,j] = m[i,j] / sum
                if str(m[i,j]) == str(float('NaN')) :
                    cli.warning('a matrix is malformed (probably frequency null for all bases)')

        return m

    
    def info_content(self):
        #
        # TODO parameter e (prob/freq problem)
        #
        
        m    = self.renormed(self.matrix)
        I, J = m.shape
        S = 0.0
        for i in range(I):
            s = 0
            for j in range(J):
                if m[i][j] != 0.0:
                    s += -m[i][j] * (log(m[i][j]) / log(2))
            S += (2.0 - s)
        return  S
        #return  S / float(I)


    #
    # OUPUT
    #
    def to_str_format(self):
        m = numpy.transpose(self.matrix)
        str = '>%s %s length=%d acc=%s\n' % (self.id, self.family, len(self), self.acc)
        for i in range(m.shape[0]):
            l = ['%6s' % ('%.2f' % a) for a in m[i]]
            str += self.bases[i] + '  ' + ' '.join(l) + '\n'
        return str


    def to_motility_format(self):
        pwm = []
        for i in range(self.matrix.shape[0]):
            pwm.append(self.matrix[i].tolist() + [0.0])

        return pwm

    def to_INCLUSIVE_format(self):
        s = ['#INCLUSive Motif Model v1.0']
        s += ['#']
        s += ['#ID = %s' % self.id]
        s += ['#W = %d' % len(self)]


        for i in range(self.matrix.shape[0]):
            str = ''
            for j in range(self.matrix.shape[1]):
                str += '%f\t' % (self.matrix[i][j])
            s += [str]

        return '\n'.join(s) + '\n'

    def to_MSCAN_format(self):
        m = numpy.transpose(self.matrix)
        str = '>%s (%s)\n' % (self.id, self.family)
        for i in range(m.shape[0]):
            str += numpy.array2string(m[i])[1:-1] + '\n'
        return str

    def to_patser_format(self):
        """
        Return a string coding the matrix in patser format
        example:
        A    |    1    3    2    0    8    0    0    0    0    0    1    2
        C    |    2    2    3    8    0    8    0    0    0    2    0    2
        G    |    1    2    3    0    0    0    8    0    5    4    5    2
        T    |    4    1    0    0    0    0    0    8    3    2    2    2
        """
        m = numpy.transpose(self.matrix)

        str = ''
        for i in range(m.shape[0]):
            str += self.bases[i] + '|'
            for j in range(m.shape[1]):
                str += '%f\t' % (m[i][j] * self.ba)
            str += '\n'
        return str


    def reverse_complement(self):
        """
        in place reverse complement
        """


        m = numpy.transpose(self.matrix)

        mref = copy(m)
        m[0,:] = mref[3,::-1]
        m[1,:] = mref[2,::-1]
        m[2,:] = mref[1,::-1]
        m[3,:] = mref[0,::-1]

        self.matrix = numpy.transpose(m)


    def to_aude_format(self):
        """
        Return a string coding the matrix in patser format
        example:
        A    |    1    3    2    0    8    0    0    0    0    0    1    2
        C    |    2    2    3    8    0    8    0    0    0    2    0    2
        G    |    1    2    3    0    0    0    8    0    5    4    5    2
        T    |    4    1    0    0    0    0    0    8    3    2    2    2
        """
        m = numpy.transpose(self.matrix)

        #print m, self.matrix

        str = '%s\t%s\n' % (self.id, self.family)
        for i in range(m.shape[0]):
            str += self.bases[i] + '|'
            for j in range(m.shape[1]):
                str += '%f\t' % (m[i][j] * self.ba)
            str += '\n'
        return str


    def to_alignment(self):
        align = [ '' * int(round(self.ba)) for i in range(len(self))]
        for i in range(len(self.matrix)):
            for j in range(4):
                align[i] += self.bases[j] * int(round(self.ba * self.matrix[i][j], 0))
        #transpose
        str = [ '' for j in range(len(align[0])) ]
        RANGE = min([len(i) for i in align])
        for l in align:
            for j in range(RANGE):
                str[j] += l[j]

        return '\n'.join(str)
        
        
    def to_weblogo(self, outdir='./logo/', bin='weblogo', type='PNG', title=None, filename=None):
        f = open('tmp.fasta', 'w')
        f.write(self.to_alignment())
        f.close()
        title = title or self.id
        filename = filename or self.id.replace('$', '\$')
        os.system('%s -F %s -Y yes -c classic --errorbars=NO -t "%s" -f tmp.fasta > %s/%s.%s' % (bin, type, title, outdir, filename, type))
        os.remove('tmp.fasta')
        
    def mean(self, a,c,g,t):
        m = self.matrix
        N = self.ba
        I, J = m.shape

        S = 0.0
        
        for i in range(I):
            S += a * (log((m[i][0]*N+a)/(N+1)) - log(a))
            S += c * (log((m[i][1]*N+c)/(N+1)) - log(c))
            S += g * (log((m[i][2]*N+g)/(N+1)) - log(g))
            S += t * (log((m[i][3]*N+t)/(N+1)) - log(t))
        return S


    def var2(self, a,c,g,t):
        m = self.matrix
        N = self.ba
        I, J = m.shape

        S = 0.0
        for i in range(I):
            S += a * (log((m[i][0]*N+a)/(N+1)) - log(a)) ** 2
            S += c * (log((m[i][1]*N+c)/(N+1)) - log(c)) ** 2
            S += g * (log((m[i][2]*N+g)/(N+1)) - log(g)) ** 2
            S += t * (log((m[i][3]*N+t)/(N+1)) - log(t)) ** 2

        T = 0.0
        for u in range(I-1):
            for v in range(u+1,I):
                X = 0.0
                X += a * (log((m[u][0]*N+a)/(N+1)) - log(a))
                X += c * (log((m[u][1]*N+c)/(N+1)) - log(c))
                X += g * (log((m[u][2]*N+g)/(N+1)) - log(g))
                X += t * (log((m[u][3]*N+t)/(N+1)) - log(t))
                Y = 0.0
                Y += a * (log((m[v][0]*N+a)/(N+1)) - log(a))
                Y += c * (log((m[v][1]*N+c)/(N+1)) - log(c))
                Y += g * (log((m[v][2]*N+g)/(N+1)) - log(g))
                Y += t * (log((m[v][3]*N+t)/(N+1)) - log(t))

                #print u,v,X,Y
                T += 2*X*Y
        print S+T
        return S + T - self.mean(a,c,g,t)**2


    def var(self, a,c,g,t):
        m = self.matrix
        N = self.ba
        I, J = m.shape

        p = [a,c,g,t]
        V = 0.0

        for i in range(I):
            S = 0.0 # E(X)
            D = 0.0 # E(X^2)
        
            for j in range(4):
                z = (log((m[i][j]*N+p[j])/(N+1)) - log(p[j]))
                S += p[j] * z
                D += p[j] * z**2
            V += D - S**2
        return V

#-------------------------------------------------------------------------------
# OBJECTS

# CODE
#-------------------------------------------------------------------------------

def to_weblogo(matrices, bin='weblogo/seqlogo', outdir='./', type='PDF', options='', suffix=''):
    for m in matrices:
        f = open('tmp.fasta', 'w')
        f.write(m.to_alignment())
        f.close()
        #title = m.id.replace('$', '\$') + suffix
        #filename = m.id.replace('$', '\$') + suffix
        title = m.MA
        filename = m.MA

        os.system('%s %s -F %s -Y -c   -n -t "%s" -f tmp.fasta > %s/%s.%s' % (bin, options, type, title, outdir, filename, type.lower()))
        os.remove('tmp.fasta')
        sys.stdout.write(title + '\n')



##########################################################################################
#
# Add by laurie 
#
##########################################################################################

def read_matrix_file(inputfile):
	"""
	Parse the fasta like file to separate the headers and the matrices

	inputfile -- a txt file in fasta like format with headers and horizontally pwm matrices
	
	return    -- a list of couples (header, matrix)
	"""
    
        try:
            fileobj = open(inputfile)  
            ignore = fileobj.read(1)

            matrices=[]

            if ignore != '>':  
                print "The first character in the file is: " + ignore  
             	print "not '>', so sys.exit() is being invoked."  
                sys.exit(1) 
                 
            text = fileobj.read()  
            records = text.split('>')
            matlist = []  
            listcount = 0

            for i in records:  
                i = '>' + i  
                matlist.append(i)  
                listcount += 1 

            len_line=0
            for mat in matlist:  
                splitCR = mat.split('\n') 
                header= splitCR[0] 
                matrice = []
                for ind in range(4):
                    if ic.verif_jaspar(splitCR[ind+1]) == False:
                        cli.error('a matrix is not in correct format: '+header)
                        sys.exit(1)
                    if ind==0:
                        len_line=len(splitCR[ind+1])
                    if len(splitCR[ind+1]) != len_line:
                        cli.error('a matrix is malformed: '+ header)
                        sys.exit(1)
                    matrice += [ map(float, splitCR[ind+1].split() )]
    	            
                matrices.append((header,matrice))	
                
            return matrices
        except ValueError:
            cli.error("A matrix in the file contains characters instead of numbers. Please fix the file before starting again.")
            sys.exit(1)

def decode_matrix_title(title):
    """
    Decode the header of personal matrix files. Deal with the optional fields

    title  -- the string title of a matrix

    return -- the fields in the title: id, name, tax_group and acc 
    """
    try:
        split = title.split(',')
        if len(split) == 1:
            print "Warning: a matrix in the file is missing information "
            id=split
            name=""
            tax = ""
            acc=""
        elif len(split) == 2:
            id, name= split
            tax = ""
            acc=""
        elif len(split) == 3:
            id,name,tax=split
            acc=""       
        else:
            id, name, tax, acc = split

        return ( id, name, tax, acc )

    except:
        cli.error("Error: the format of a title in the fasta file is not correct. Please check your file")
        raise
        sys.exit(1)
        
	    
########################################################################################################

########################################
#                                      #
#    XML INPUT/OUPUT                   #
#                                      #
########################################
def load_matrices_info(filename):
    """
    filename -- xml matrices info file
    
    Return dict indexed by matrix id of dict {'family': ***, ...}
    """
    h = {}

    d = xmltramp.parse(open(filename).read())
    for matrix in d:
        h[matrix('id')] = matrix()
    return h
    
    
def matrices2xml(f, matrices):
    out = open(f, 'w')
    out.write ('<?xml version="1.0" encoding="ISO-8859-1"?>\n')
    out.write ('<matrices db="TRANSFAC" version="6.0">\n')

    i = 0
    for matrix in matrices:
        params = { 'ref'    : i,
                   'id'     : saxutils.escape(matrix.id),
                   'family' : saxutils.escape(matrix.family),
                   'length' : len(matrix),
                   'desc'   : saxutils.escape(matrix.desc)
                   }
                   
        out.write('\t<matrix ref="%(ref)d" id="%(id)s" family="%(family)s" length="%(length)d" desc="%(desc)s"/>\n' % params)
        i += 1

    out.write ('</matrices>\n')
    out.close()


#-------------------------------------------------------------------------------
# CODE




