####################
# version 2.0
####################


################################################################################
#                                                                              #
# NAME Core.Types.sl 
# DESC Structured List Store and read list from structured text files 
#                                                                              #
################################################################################
"""
:NAME:
Core.Types.sl

:DESC:
Structured TEXT List Reader/Writer

:LONG_DESC:
Read and Store list with meta information in text format

:EXAMPLE:
+++++++++
> title
%{comment name1} value
%comment_name2   value
# skipped comment
element1
element2
+++++++++

:TODO:
SLReader.next() : optimise

:VERSION:
0.1
2006-01-23

:AUTHOR:
Matthieu Defrance

:LICENSE:
GPL
"""

# IMPORTS
#-------------------------------------------------------------------------------
import os
import re
import StringIO


import Core.Types.string as string
import Core.IO.utils as io
from Core.Types.simple import Struct

#-------------------------------------------------------------------------------
# IMPORTS

# CODE
#-------------------------------------------------------------------------------

def f2sl(f):
    return SLReader(f).next()

def sl2uf(sl, f):
    w = SLWriter(f)
    w.append(sl)

def uf2sl(f):
    f = io.uopen(f, 'U') 
    try:
        return SLReader(f).next()
    except StopIteration:
        return SL([])

def string2sl(s):
    f = StringIO.StringIO(s)
    return s2sl(f)

class SL(list):
    """
    Structured List Object
    attributes:
        title  string
        info  dict
    
    """
    def __init__(self, *args):
        """
        default list *args
        """
        list.__init__(self, *args)
        
        self.title    = ''
        self.info     = {}
    
    def from_file(f):
        """Use first structured list in file
        """
        return SLReader(f).next()
    
    from_file = staticmethod(from_file)
    
    
    def get_info(self):
        return self.info
    
    def __repr__(self):
        return '%s' % self.title + `self.info` + `list(self)`


class SLReader:
    """
    Structured List Reader Iterator

    """
    def __init__(self, f):
        """
        f -- file or filename
        """
        self.f = f
        self.item = None

    def __iter__(self):
        return self

    def next(self):
        """Read next structured List in file
        return SL Object
        """
        comment = {}
        l = []
        title = 'no name'
        item  = False
        eof   = False
        
        # begin hawfull loop
        while 1:
            #z = self.f.readline()

            line = self.f.readline()
            if not line:
                eof = True
                break
            if line.startswith('>'):
                if not item:
                    title = line[1:].strip()
                    item = True
                    continue
                else:
                    self.f.seek(-len(line), 1)
                    break

            line = line.strip()
            if line == '':
                continue
            if line[0] == '#':
                continue
            if line[0] == '%' and item:
                m = re.match(r'%\{(.*?)\} (.*)', line)
                if m:
                    comment[m.group(1).strip()] = m.group(2).strip()
                    continue
                
                m = re.match(r'%(\w+)\s+(.*)', line)
                if m:
                    comment[m.group(1)] = m.group(2).strip()
                    continue
                continue
        
            l.append(line.split('#', 1)[0].rstrip())

        if l == [] and eof:
            raise StopIteration
        
        
        r = SL(l)
        r.title = title
        r.info = comment
        return r


class SLWriter:
    """
    Structured List writer

    """
    def __init__(self, f):
        """
        f -- file or filename
        """
        self.f = io.uopen(f, 'w')

    def append(self, sl):
        """Append structured list to file
        sl -- SL
        """
        self.f.write('> %s\n' % sl.title)
        for key, value in sl.info.items():
            self.f.write('%%{%s} %s\n' %  (key, value))
        for elt in sl:
            self.f.write(str(elt) + '\n')
	self.f.flush()


########################################
#                                      #
#               GABGET                 #
#                                      #
########################################

def list_dir(dir, ext=''):
    """Return list of sl found in directory
    """
    l = []
    for f in os.listdir(dir):
        if not os.path.isfile(os.path.join(dir,f)) and utils.suffix(f) != ext:
            continue
        sl = f2sl( open(os.path.join(dir,f)) )
        l.append(sl)
    return l


class SL_DIR:
    """Store all structured lists in a directory
    
    access to list is made by sub attributes l
        
    """
    
    def __init__(self, dir, ext=''):
      """Scan directory for list file
      dir       -- directory to scan
      ext       -- only search for this extension [optional]
      """
      
      self.l = Struct()
      
      for sl in list_dir(dir, ext):
          setattr(self.l, string.utils.convert2attr(sl.title), sl)    


    def __repr__(self):
        return '\n'.join([`i` for i in self.l.__dict__.values()])



    









#-------------------------------------------------------------------------------
# CODE
