####################
# version 2.0
####################


################################################################################
#                                                                              #
# NAME Stats.utils  
# DESC Statistics utilities
#                                                                              #
################################################################################
"""
:NAME:
Stats.utils

:DESC:
Statistics utilities for discrete statistics

"""
# IMPORTS
#-------------------------------------------------------------------------------
from copy import copy
from math import *
import Core.types as types
#-------------------------------------------------------------------------------
# IMPORTS

# CODE
#-------------------------------------------------------------------------------



def test_pmf(d):
    return types.list.sum(d.values())

def shifted_pmf(d, shift=None):
    shift = shift or -min(d.keys())

    dr = {}
    for v in d:
        dr[v+shift] = d[v]
    return dr

def pmf_change_values(d, f):

    dr = {}
    for v in d:
        dr[f(v)] = dr.get(f(v), 0.0) + d[v]
    return dr


def auto_compressed_pmf(d, nbin=10):
    rate = float(nbin) / len(d)
    assert(rate <= 1.0)
    return rate, compressed_pmf(d, rate)

def compressed_pmf(d, rate):
    """
    compress given distribution

    d MUST have values in [0, ...]

    """
    dr = {}

    for v in d:
        cv = int(v*rate)
        dr[cv] = dr.get(cv, 0.0) + d[v]

    return dr


def expand_pmf(d, rate):
    """
    d -- a compressed pmf
    """
    dr = {}
    for v in d:
        vrange = ( int(v/rate), -1 + int((v+1)/rate) )
        dr[vrange] = d[v]

    return dr    

    
def pmf(values):
    """Return probability mass function
    dict key=value, value=prob
    
    """
    count = {}
    for v in values:
        count[v] = count.get(v, 0) + 1

    N = float(len(values))
    for v in count:
        count[v] /= N

    return count


def pmf_tail(d, pvalue, lowerValue=0):
    """
    keep only distribution corresponding to tail >= pvalue
    group other values on 0
    
    """
    v, p = types.dict.sorted_keys_values(d)    
    i = len(p)
    pv = 0.0
    while i >  0 and pv < pvalue:
        i -= 1    
        pv += p[i]
    
    if v[i] <= 0:
        raise Error

    dr = {lowerValue: 1.0 - pv + p[i]}
    for j in range(i+1, len(v)):
        dr[v[j]] = p[j]

    return dr
    
def cdf(pmf):

    cdf = {}
    #v, p = types.dict.sorted_keys_values(pmf)
    cp = 0.0
    for v,p in zip(*types.dict.sorted_keys_values(pmf)):
        cp += p
        cdf[v] = cp
    return cdf

def cdf_upper(pmf):
    """
    p(x) : P[X>=x]
    """
    cdf = {}
    cp = 1.0
    for v,p in zip(*types.dict.sorted_keys_values(pmf)):
        cdf[v] = cp
        cp -= p
    return cdf



def probs(d, a=None, b=None, lowerProb=None):
    """
    return list of prob in range (a,b+1) 
    default : a,v = min(d.keys()), max(d.keys())
    
    d -- distribution dict
    """
    a = a or min(d.keys())
    b = b or max(d.keys())
    x = []
    for i in range(a, b+1):
        x.append(d.get(i, 0.0))
    return x


def pmf_tail_renormed(d, percent=1.1):
    """
    **** TFM-E balance ****

    d -- tail distribution
    """
    dr = {}
    for k in d:
        if float(k) == 0.0:
            dr[0] = 1.0 - (1.0 - d[0]) * percent 
            if dr[0] < 0.0:
                raise RenormError
        else:
            dr[k] = d[k] * percent

    return dr

def pvalue2value(d, pvalue):
    """
    TODO error case : can not find value P(value) >= pvalue
    """
    v, p = types.dict.sorted_keys_values(d)

    i = len(p)
    pv = 0.0
    while i > 0 and pv < pvalue:
        i -= 1
        pv += p[i]

        
    #print pv - p[i], pv
    #print '<'
    #print types.list.sum(p[i+1:])
    #print v[i+1]
    #print v[min(i+1, len(v)-1)]
    #print '>'
    return v[min(i+1, len(v)-1)]


def value2pvalue(d, value):
    v, p = types.dict.sorted_keys_values(d)
    pv = 0.0
    i = len(p)
    while i > 0 and v[i-1] >= value:
        i -= 1
        pv += p[i]
        
    return pv

factorial = lambda n: n and factorial(n-1)*long(n) or 1

########################################
#                                      #
#        Poisson
#                                      #
########################################
def poissonp(x, mu):
    S = 0
    for i in range(x):
        S += exp(-mu) * mu ** i / factorial(i)
    return 1.0 - S

def sum_of_poisson(lamb, i, j):


    if lamb <= 0 and (i > 0 or j > 0):
        print 'Error'
        return None

    logLamb = log(lamb)

    if i == 0:
        S = exp(-lamb)
    else:
        S = 0

    if j == 0:
        logPois = -lamb
    else:
        logPois = -lamb + logLamb
        if i <= 1:
            S += exp(logPois)

    if i <= 1:
        start = 2
    else:
        start = i
        for x in range(2, start):
            logPois += logLamb - log(x)


    for x in range(start, j+1):
        logPois += logLamb - log(x)
        S += exp(logPois)
        if x > lamb and exp(logPois) <= 0:
            break
        prev = S
    return S


########################################
#                                      #
#         Generic functions
#                                      #
########################################

def multi(x, func, *args):
    """
    do y = func(x, *args) for each x
    return list of y if x is list or y 
    """
    if type(x) == list:
        return [func(i, *args) for i in x]    
    else:
        return func(x, *args)


def mean(values):
    return types.list.sum(values) / float(len(values))

def var(values):
    m = mean(values)
    return types.list.sum( [(x-m)**2 for x in values] ) / float(len(values)-1)


def mean_var(values):
    """
    return tuple mean and var
    """
    return mean(values), var(values)

def table(v, binFunc=None):
    """Count table
    v       -- values
    binFuc  -- binarize function default=round
    return tuple (values, counts)
    """
    binFunc = binFunc or (lambda x: round(x))
    v = map(binFunc, v)
    #v = map(float, v)
    t = {}
    for i in v:
        t[i] = t.get(i, 0) + 1
    l = t.items()
    l.sort()
    return [i[0] for i in l], [i[1] for i in l]

########################################
#                                      #
#         
#                                      #
########################################

def thdist(x_range, step, func, *params):
    """x,y for theorical distribution
    step -- number of point per unit
    example thdist( (-1.0,1.0), 2, norm, a, b) 
    return x,y
    """
    x_range = [i/float(step) for i in range(int(x_range[0]*step), int(x_range[1]*step+1))]
    return x_range, multi(x_range, func, *params)


########################################
#                                      #
#         
#                                      #
########################################

    
def reshape(l, lr, niMin=5):
    """
    Reshape lists l and lr to match value(l[i]) >= niMin for each i
    
    Return new l and lr
    """
    l  = copy(list(l))
    lr = copy(list(lr))

    n    = float(reduce(lambda a,b : a+b, l) )
    n_lr = float( reduce(lambda a,b : a+b, lr) )

    if n_lr == 0.0:
        return [], []

    pr = [i / n_lr for i in lr]
    for i in range(len(l)):
        if pr[i]*n >= niMin or i >= len(l) -1: 
            break

    if i != 0:
        l[i] = reduce(lambda a,b : a+b, l[:i+1])
        lr[i] = reduce(lambda a,b : a+b, lr[:i+1])
        del (l[:i])
        del (lr[:i])
        del (pr[:i])

    t = range(len(pr))
    t.reverse()
    for i in t:
        if pr[i]*n >= niMin and i <=  len(l) -1:
            break

    if i != len(l):
        l[i] = reduce(lambda a,b : a+b, l[i:])
        lr[i] = reduce(lambda a,b : a+b, lr[i:])
        del (l[i+1:])
        del (lr[i+1:])

    return l, lr


#-------------------------------------------------------------------------------
# CODE
