####################
# version 2.0
####################



"""
Toolbox for scanning background models and sequences, and outputting the results

methods:
    hit_frequency
    hits_scan
    hits_scan_fast
    scan_
    scan
    dictlist2str
    hits2txt
    r2txt
    r2txt_details
    correlations
    get_correlation
    get_proba
    get_intersection
    get_overlap
"""

import time
import sys
from numpy import zeros
import math

import Core.Types.tab
import Core.Types.list
import Core.UI.cli as cli

import Bio.hit
import Bio.sequence
import TFBSscanner
import tab
import scan as mod_scan
import window
import pvalue
import outputXML
import os
from operator import itemgetter




def hit_frequency(h, i_m, location=None):
    """
    positionnal hit count frequency table
    """
    location = location or h.sequences.location
    T_k = zeros( location[1] - location[0] + 1 , 'i')
    T_lambda = None #float array    
    
    T_k += hit.expand_many(hits[i_m,:], location) 

    # build T_lambda
    T_lambda = Core.Types.tab.smooth(T_k/float(len(h.sequences)), SMOOTHING_WINDOW_WIDTH)

    return T_lambda


def hits_scan(sequences, matrices, bgdb):
    """
    sequences      -- list of Sequences
    
    matrices       -- list of Matrices
    
    bgdb           -- BGDB
    
    return         -- Bio.types.hit.Hits
    """

    h = Bio.hit.Hits(sequences, matrices)
    M = matrices
    S = sequences
    scanner = bgdb.scanner

    info = cli.Info(len(S)*len(M), showTime=1)
    for i_m in range(len(M)):
        for i_s in range(len(S)):
            bg = bgdb.get(S[i_s],M[i_m])
            h[i_m,i_s] = scanner.scan([S[i_s]], M[i_m], bg)[0]
            info('scan hits')

    return h

def hits_scan_fast(sequences, matrices, bgdb):
    """
    scan background model

    sequences      -- list of Sequences
    
    matrices       -- list of Matrices
    
    bgdb           -- BGDB
    
    return         -- Bio.types.hit.Hits
    """
    # 1. sort sequences given their bg group
    get_group = lambda o: bgdb.get(S[i_s],M[i_m])
    
    
    
    h = Bio.hit.Hits(sequences, matrices)
    M = matrices
    S = sequences
    scanner = bgdb.scanner

    
    info = cli.Info(len(M), showTime=1)
    for i_m in range(len(M)):

        f = lambda o: bgdb.get(o,M[i_m])

        for (bg,  seqIndices) in Core.Types.list.categorize_idx(sequences, f).items():

            r = scanner.scan([S[i] for i in seqIndices], M[i_m], bg)

            for j in range(len(seqIndices)):
                h[i_m, seqIndices[j]] = r[j]
		
        info('Scanning hits')

    return h



def scan_(sequences, matrix, hits, bgModels, parameters=None, numberOfScannedMatrices=1):
    """
    scan the user sequences and extract interesting windows according to results from scan of bg

    hits       -- hits for one matrix
    
    parameters -- dict of parameters keys=(minSize, maxSize)
    
    bgModels   -- list of Bg
    
    return     -- list of Window Object
    """
    t = tab.Tab(sequences)
    t.build(hits, bgModels, parameters)

    score = mod_scan.scan(t.T_k, t.T_alpha, t.T_alpha_b, parameters)

    r = mod_scan.extract_growing_regions(score, parameters['minSize'], parameters['maxSize'])

    # append windows
    wl = []
    for i,j in r:
        a, b = tab.indicies2location((i,j), t.location)
        pv = pvalue.pvalue(t.T_k, t.T_alpha_b, i,j)
        ev = pvalue.evalue(sequences.location, parameters['minSize'], parameters['maxSize'], pv, numberOfScannedMatrices)
        w = window.Window(a, b, matrix, pv, ev) 
        w.score = score[i:j].tolist()
	if parameters['pvalmax'] != "" :
		if pv<parameters['pvalmax']:
	        	wl += [ w ]
	else:
		wl += [ w ]
    return wl


def scan(sequences, matrices, bgdb, output, parameters=None):
    """
    launch scan of background model and sequences.
    format the hits and interesting windows
    
    sequences      -- list of Sequences
    
    matrices       -- list of Matrices
    
    bgdb           -- BGDB
    
    output         -- directory where to put the results files
    
    parameters     -- user parameters for the scan
    
    return         -- a vector with parameters and result windows
    """
    parameters = parameters or {'minSize': 30, 'maxSize': 1000, 'ratio': 2.5, 'p': 0.0, 'targetSize': 200, 'top':10}


    ids = Core.Types.list.intersection([matrix.id for matrix in matrices], list(bgdb.idx.list_values('matrixId')))
    #if len(matrices) != len(ids):
    #    cli.Warning('Not all matrices can be found into bg model')

    matrices = [m for m in matrices if m.id in ids]

    location = Bio.sequence.find_location(sequences)
    
    # find all hits
    h = hits_scan_fast(sequences, matrices, bgdb)
    
    wl = window.WindowList(title=str(sequences), maxSize=parameters['top'])
    info = cli.Info(len(matrices), 1)
    evc = 0
    for i_m in range(len(matrices)):
        info('Scanning matrix %s' % `matrices[i_m]`)
        bgModels = [bgdb.get(s, matrices[i_m]) for s in sequences]
        l = scan_(sequences, matrices[i_m], h[i_m,:], bgModels, parameters, numberOfScannedMatrices=len(matrices)) # calculate score function and extract windows
        wl += l
        evc += len(wl)

    #for w in wl:
       # w.evalue = w.pvalue * evc * len(matrices) wrong

    wl.set_info(h)
    

    ########################################################################################
    # select the hits of all the matrices in the extracted windows: Needed for visualisation
    ########################################################################################  
    hitsToSave={}
    for m in range(len(h.matrices)):

        for s in range(len(h.sequences)):

            if len(h[m,s]) > 0 and len(h[m,s][0]) > 1:

                for x in h[m,s]:

                    #for start in wl:

                        startHit=x[0]+h.sequences[s].location[0]
                        word=h.sequences[s].sequence[x[0]:x[0]+len(h.matrices[m])]
                        #endHit=x[0]+h.sequences[s].location[0]+len(h.matrices[m])

			            # select the hits that are in at least one of the extracted windows
                       # if (endHit > start.i and endHit < start.j) or (startHit > start.i and startHit < start.j): 
                        keyHit=(h.matrices[m].id,h.sequences[s].id,startHit) # key is here to avoid saving twice the same hit
                        hitsToSave[keyHit]=[x[1],len(h.matrices[m]),word]
    keys=hitsToSave.keys()
    keys.sort(key=itemgetter(2),reverse=False)
    keys.sort(key=itemgetter(0),reverse=False)
    keys.sort(key=itemgetter(1),reverse=False)
    finalHits=[]
    for k in keys:
	    finalHits.append([k,hitsToSave[k]])	
	
    

    #####################################
    # prepare a vector with the results
    #####################################
    r = {
        'R':  [i.to_dict() for i in wl],
        'matricesSize'  : len(matrices),
        'sequencesSize' : len(sequences),
        'sequencesName' : sequences.name,
	'sequences'	: [s.id for s in sequences],
        'start'         : location[0],
        'end'           : location[1],
        'parameters'    : parameters,
        'date'          : time.ctime(),
            }

    

 
    return (r,finalHits)

# template for outputting results on screen
#;rank,matrixID,TF,info,gc, Location/TSS,  P-value, E-value, nhits, sequences
HEADER1 = '''
;TFM Explorer'''

HEADER2 = ''';Date              : %(date)s
;Scanned location  : %(start)s:%(end)s
;Scanned sequences : %(sequencesName)s (%(sequencesSize)d sequences)
;Scanned matrices  : (%(matricesSize)d matrices)
;Parameters       : %(parameters)s
;
; column headers
;    1    rank               window rank
;    2    matrix             matrix name
;    3    tf                 transcription factor name
;    4    info               information of content of the matrix 
;    5    gc                 gc content of the matrix (G+C)%%
;    6    location           location of the window (relatively to TSS)
;    7    pvalue             pvalue of the window
;    8    nh                 number of motif occurrences in the window
;    9    ns                 percent of sequence that have at least one motif occurrence in the window
;
;rank\tmatrix\ttf\tinfo\tgc\tlocation\tpalue\tnh\tns'''


WINDOW_FORMAT = '%(matrixId)-18s\t%(matrixName)-18s\t%(matrixInfo)5.2f\t%(matrixGC)4.2f\t[%(i)+05d:%(j)+05d]\t%(pvalue).3e\t%(numberOfHits)4d\t%(percentOfSequences)3d%%'


def dictlist2str(l, FORMAT=WINDOW_FORMAT):
    """
    Transforms a dictionnary with the results for a window in a formatted string
    
    l      -- a dictionnary of information
    FORMAT -- A string pattern
    
    return -- a string with all the information of a result window
    """
    s = []
    for i in range(len(l)):
        w = FORMAT % l[i]
        s += ['%2d ' % (i+1) + w]
    return '\n'.join(s)

def hits2txt(l, sequences, matrix):
    """
    Transforms a list of Hits into a string
    
    l         -- list of hits
    
    sequences -- list of sequences
    
    matrix    -- a weight matrix
    
    return    -- a string with the details of each hit
    """
    r = []
    for i in range(len(l)):
        if len(l[i]) == 0:
            continue
        #r += [str(i[0][0]) + '\n\t' + ' '.join(['%+05d' % h[1] for h in i])]
        for h in l[i]:
            s = sequences[i]
            word = s.sequence[h[1]-s.location[0]:h[1]-s.location[0]+len(matrix)]

            if h[2] == '-':
                word = Bio.sequence.reverse_complement(word)
            r += [ '%s\t %+05d %s %5.2f %s' % (h[0], h[1], h[2], h[3], word)]

    return '\n'.join(r)

def r2txt(r, info=None):
    """
    Transforms the result of TFME in a simple text to output
    
    r      -- dictionnary with the results of TFME
    
    info   -- complementary information
    
    return -- a string with the information
    """
    R =  [ HEADER1 ]
    if info:
        R += [ info ]
    R += [ HEADER2 % r]
    R += [ dictlist2str(r['R'], WINDOW_FORMAT) ]
    return '\n'.join(R)

def r2txt_details(r, sequences=None, matrices=None):
    """
    Transforms the result of TFME in a text ( with details of each window) to output
    
    r         -- dictionnary with the results of TFME
    
    sequences -- list of sequences
    
    matrices  -- list of matrices
    
    return    -- a string with the detailed information
    """
    #return
    R = []
    R += [';']
    R += [';  Details']
    R += [';']
    for i in range(len(r['R'])):
        x = r['R'][i]
        matrix = Core.Types.list.select_first(matrices, lambda x: x.id == r['R'][i]['matrixId'])
        R += ['%2d %s length=%d %d hits %d sequences' % (i+1, matrix.id, len(matrix), x['numberOfHits'], x['numberOfSequences']) ]
        #R += ['%s' % str(matrix) ]
        R += [ '\n'.join(x['hitsFull']) ]
        #R += [ hits2txt(x['hitsFull'], sequences, matrix) ]
        R += ['']

    return '\n'.join(R)


def correlations(r):
    """
    calculation of correlation coefficients and overlapping sequences between hit windows, 2 by 2.

    r      -- results of tfme, list of windows with their sequences, matrices...
    
    return -- a list of correlations for all the windows.
    """
    corr=[]

    nbWindows=len(r['R'])
    for i in range(nbWindows-1):
        for j in range(i+1,nbWindows):

            pairWise_corr={}
            correlation = round(get_correlation(r, i, j), 4)
            if correlation > 0.0:
                intersection = len(get_intersection(r, i, j))
                ov = get_overlap(r, i, j)

                if ov.has_key('location'):
                    overlapping = {'location' : '[%+05d:%+05d]' % (ov['location'][0], ov['location'][1]),
                                      'o' : '(%.2f%%)' % ( (ov['p1'] + ov['p2']) / 2.0 )}
                else:
                #overlapping = {'location' : '', 'o':'', 'r' : '%.2f%%' % ( (ov['r1'] + ov['r2']) / 2.0 )}
			        overlapping = {'location' : '', 'o':''}

                pairWise_corr['matrix1']=r['R'][i]['matrixId']
                pairWise_corr['matrix2']=r['R'][j]['matrixId']
                
                pairWise_corr['factor1']=r['R'][i]['matrixName']
                pairWise_corr['factor2']=r['R'][j]['matrixName']

                pairWise_corr['location1']=r['R'][i]['locationStr']
                pairWise_corr['location2']=r['R'][j]['locationStr']

                pairWise_corr['correlation']=correlation
                pairWise_corr['intersection']=intersection
                pairWise_corr['overlapping']= overlapping
			
                corr.append(pairWise_corr)

    return corr

def get_correlation(r, i, j):
	"""
	calculation of correlation coefficient between two result windows

	r      -- results of tfme: list of hit windows
	
	i      -- index of the first window to use
	
	j      -- index of the second window to use
	
	return -- the correlation value
	"""
       # return Core.Types.list.correlation(r['R'][i]['sequencesVector'], r['R'][j]['sequencesVector'])
   
        vect1=r['R'][i]['sequencesVector']
        vect2=r['R'][j]['sequencesVector']
        
        I=0
        vals=[0,1]
        for a in vals:
            for b in vals:
                try:
                    div=get_proba(vect1,a,vect2,b)/(get_proba(vect1,a)*get_proba(vect2,b))
                    log=math.log(div,2)
                    sub=get_proba(vect1,a,vect2,b)*log
                except:
                    sub=0
                    
                I+=sub
                
        return I   

def get_proba(vect1,val1,vect2=None,val2=None):
    """
    Calculation of probability of x:x between two sequence vectors (used in correlation calculation)
    
    vect1  -- a sequence vector (ex: [0,1,...,1,0] with 0 if no hit for the facor, 1 if at least one hit)
    
    val1   -- a value to find probability in vect 1 (0 or 1)
    
    vect2  -- a sequence vector
    
    val2   -- a value to find probability in vect 2 (0 or 1)
    
    return -- a probability (float between 0 and 1)
    """

    if vect2 is not None:
        assert (len(vect1)==len(vect2)), "Sequences vectors different in correlation calculation"

        nb=0

        for i in range(len(vect1)):
            if vect1[i]==val1 and vect2[i]==val2:
                nb+=1
        
        p=float(nb)/float(len(vect1))
        
    else:
        nb=0
        for i in range(len(vect1)):
            if vect1[i]==val1:
                nb+=1
        
        p=float(nb)/float(len(vect1))
        

    return p

def get_intersection(r, i, j):
	"""
	calculation of the intersection between two result windows

	r      -- results of tfme: list of hit windows
	
	i      -- index of the first window to use
	
	j      -- index of the second window to use
	
	return -- the intersection of the two windows
	"""
        return Core.Types.list.intersection(r['R'][i]['sequencesIn'], r['R'][j]['sequencesIn'])

def get_overlap(r, i, j):
	"""
	calculation of the overlapping parts and TFBS between two result windows

	r -- results of tfme: list of hit windows
	
	i -- index of the first window to use
	
	j -- index of the second window to use
	
	return -- a dictionnary with the overlap sequences and hits between the two windows
	"""
        overlappingLocation = Core.Types.interval.intersection(r['R'][i]['location'], r['R'][j]['location'])
        location1 = r['R'][i]['location']
        location2 = r['R'][j]['location']

        X = {}
        m1, m2 = r['R'][i]['matrixId'], r['R'][j]['matrixId']

        #X['r1'] = round(objects['hits_overlapping'].get( (m1, m2), 0.0)*100.0, 2)
        #X['r2'] = round(objects['hits_overlapping'].get( (m2, m1), 0.0)*100.0, 2)


        if overlappingLocation:
            h1 = r['R'][i]['hitsList']
            h2 = r['R'][j]['hitsList']
            ml1 = r['R'][i]['matrixLen']
            ml2 = r['R'][j]['matrixLen']
            (o1,c1), (o2, c2) = Bio.hit.overlapping(h1, ml1, h2, ml2, overlappingLocation)
            X['location'] = overlappingLocation
            X['o1'] = o1
            X['c1'] = c1
            X['o2'] = o2
            X['c2'] = c2

            X['p1'] = round(100.0 * o1 / float(max(1,c1)), 2)
            X['p2'] = round(100.0 * o2 / float(max(1,c2)), 2)
            X['locationp1'] = round(100.0 * (overlappingLocation[1]-overlappingLocation[0]+1) / float(location1[1] - location1[0] + 1), 2)
            X['locationp2'] = round(100.0 * (overlappingLocation[1]-overlappingLocation[0]+1) / float(location2[1] - location2[0] + 1), 2)
            #[%+05d:%+05d] (%.2f%%)

        return X

if __name__ == '__main__':
    main()






