#! /usr/bin/env python

###################
# version 2.0
###################

#############################################################################################
# tfmebg
# Program that takes dump files of matrices and sequences, and create a background model,
# saved into a binary file. 
#
# author:Matthieu DeFrance, modified by Laurie Tonon
# date: 28/10/2009
#
#############################################################################################

import os, sys
PYTHONPATH = [['TFME'], ['TFME', 'MLib.zlib']]
if sys.version_info < (2, 3): 
    print >> sys.stderr, 'python >= 2.3 is required!'; sys.exit()
[sys.path.insert(1, os.path.join(sys.path[0], *path)) for path in PYTHONPATH]


import imp

#********************************************************************
# Check dependencies
#********************************************************************
try:
    import numpy
    version= numpy.__version__

except ImportError:
    print 'WARNING: TFM-Explorer requires Numpy '
    sys.exit(1)
    
try:
    import Pyrex

except ImportError:
    print 'WARNING: TFM-Explorer requires Pyrex '
    sys.exit(1)


import optparse

import Core.UI.cli as cli
import Core.Persistence.op as op
import Core.IO.utils
import Bio.sequence
import pataud

import shelve



usage = """
      python  %prog --matrices=file.mat.db --sequences=file.seq.db --output=file.bg [--prioriprob=x]
"""

parser = optparse.OptionParser(usage)
parser.add_option("-m", "--matrices", action="append", dest="matrices", type="string", default=[], help="list of matrices DB files to use (must be dump files produced by data2dump.py)")
parser.add_option("-p", "--prioriprob", dest="p", help="piori probability: probability as a threshold of acceptance for pwm (default=0.001)", action="store", type="float", default=0.001)
parser.add_option("-o", "--output", dest="output", action="store", type="string", help="The background file produced.")
parser.add_option("-s", "--sequences", action="append", dest="sequences", type="string", default=[], help="list of sequences DB filenames (must be dump files produced by fasta2dump.py")
parser.add_option("-v", "--verbosity", action="store", dest="verbosity", type="int", default=1, help="display complete process on screen (between 1 and 3) (default=1)")
parser.add_option("--debug", dest="debug", action="store_true", default=False, help="show program errors. For developpers only")

(options, args) = parser.parse_args()
cli.VERBOSITY = options.verbosity

def run(args, options):
    
    # buffers for errors
    error_found = False
    error_messages = []
    
#******************************************************************
# Process user inputs
#******************************************************************

    if not options.matrices:
        cli.error('No matrices provided. Cannot continue.')
        sys.exit(1)
        
    if not options.sequences:
        cli.error('No sequences provided. Cannot continue')
        sys.exit(1)
        
    if not options.output:
        cli.error("No output file provided. Cannot continue.")
        sys.exit(1)
        
    elif os.path.dirname(options.output) != '' and not os.path.isdir(os.path.dirname(options.output)): # check if the output directory exists
        cli.error('The output directory does not exists.')
        sys.exit(1)
    else:
        if os.path.isfile(options.output): # if a dump file with the same name already exists, try to replace it
            try:
                os.remove(options.output)
            except OSError:
                cli.error("A dump file with the same name as "+options.output+" already exists and cannot be removed. Please delete or rename this file and lauch tfmebg.py again.")
                sys.exit(1)
        elif os.path.isfile(options.output+'.db'):
            try:
                os.remove(options.output+'.db')
            except OSError:
                cli.error("A dump file with the same name as "+options.output+'.db'+" already exists and cannot be removed. Please delete or rename this file and lauch tfmebg.py again.")
                sys.exit(1)        
    
    ################################################
    # read, check and prepare the sequences
    ################################################

    sequences_small=[]
    for filename in Core.IO.utils.super_glob(options.sequences):
        if os.path.isfile(filename):
            #sequences_small.append(op.oload(filename) )
            dump=op.P(filename)
            if dump.get('_info') != 'SEQUENCES':
                cli.warning('file '+ filename+' does not contain sequences. Ignored.')
            else:
                sequences_small.append(dump)
        elif os.path.isfile(filename+'.db'):
            dump=op.P(filename+'.db')
            if dump.get('_info') != 'SEQUENCES':
                cli.warning('file '+ filename+'.db'+' does not contain sequences. Ignored.')
            else:
                sequences_small.append(dump)
        else:
            cli.warning("file "+filename+" cannot be found. Ignored.")
    
    sequences_all=[]

    for seqKey in sequences_small[0].keys():
        
        if seqKey.find('%') == -1:
            sequences_all.append(sequences_small[0].get(seqKey))
    		
    sequences=Bio.sequence.Sequences(sequences_all)


    if len(sequences) == 0:
        error_found=True
        error_messages.append("Sequences set is empty.")        

    ######################################
    # read,check and prepare the matrices
    ######################################

    mDB=[]
    for filename in Core.IO.utils.super_glob(options.matrices):
        if os.path.isfile(filename):    
            #mDB.append(op.oload(filename))
            dump=op.P(filename)
            
            if dump.get('_info').find('MATRICES') == -1:
                cli.warning("file "+ filename+ "does not contain matrices. Ignored")
            else:
                mDB.append(dump)
        elif os.path.isfile(filename+'.db'):
            dump=op.P(filename+'.db')
            if dump.get('_info').find('MATRICES') == -1:
                cli.warning('file '+ filename+'.db'+' does not contain matrices. Ignored.')
            else:
                mDB.append(dump)
        else:
            cli.warning("file "+filename+" cannot be found. Ignored.")
    
    matrices = Core.Types.list.flatten([d.values() for d in mDB])
    
    if len(matrices) == 0:
        error_found=True
        error_messages.append("Matrices set is empty.")        


#***************************************************
# output possible errors
#**************************************************
    if error_found:
        print "\n[ERRORS]The following errors have been found: "
        for m in error_messages:
            print m
        sys.exit(1)



#***************************************************    
# launch background building
#***************************************************
    print "TFM-Explorer background builder job running..."
    pataud.build_BG(sequences, matrices, options.p, options.output, overlap=False, gcRange=(0.00, 1.00))
	
    print "TFM-Explorer background builder job successfully ended"

if __name__ == '__main__':
    try:
        run(args, options)
    except:
        cli.error('Error while running TFM-Explorer background builder')
        if options.debug:
            raise


